import 'dart:convert';
import 'dart:io';
import 'package:mime/mime.dart';
import 'package:alm/alm.dart';

typedef kApiMethod = Future<dynamic> Function();
typedef kRouteMethod = dynamic Function(HttpRequest);

abstract class _Api {
  HttpRequest request;

  Map<String, kApiMethod> allows = {};
  Map<String, kApiMethod> blocks = {};

  Future<void> init();

  String get headerToken => request?.headers?.value('token');

  Future<bool> headerCheckToken();
}

class Api implements _Api {

  static Function(dynamic) log=(o){print('Api $o');};

  static String generateToken(String pass, {Duration duration}) => Alm.tokenGen(pass, duration: duration);

  static bool expiredToken(String token, {Duration duration}) => Alm.tokenExpired(token);

  static Map decodeToken(String token) => Alm.tokenDecode(token);

  dynamic postData;
  Map<String,dynamic> postJson;
  Map<String,dynamic> postFiles;
  Map user = {};

  bool get isGet => method == 'GET';

  bool get isPost => method == 'POST';

  String get method => request.method ?? 'GET';

  bool get isBoundary => contentType.parameters.containsKey('boundary');

  ContentType get contentType => request.headers.contentType;

  List<String> get arguments => request.uri.pathSegments;

  @override
  Future<void> init() {}

  Map get requestInfo {
    dynamic map = {};
    map['ip'] = '${request.connectionInfo.remoteAddress.host}:${request.connectionInfo.remotePort}';
    map['header'] = <String, dynamic>{};
    request.headers.forEach((name, values) {
      map['header'][name] = values.length > 1 ? values.length : values.first;
    });
    map['time'] = Alm.timestamp();
    return map;
  }

  Future<String> enter(HttpRequest request) async {
    this.request = request;
    var pathSegments = request.requestedUri.pathSegments;
    var action = pathSegments.last;
    dynamic res;
    postJson={};
    postFiles={};
    if (isPost && contentType.mimeType.toLowerCase() == ContentType.json.mimeType.toLowerCase()) {
      postJson = Map.from(Alm.map(await utf8.decoder.bind(request).join()));
    }

    if (blocks.containsKey(action)) {
      if (await headerCheckToken()) {
        res = Alm.error('[$runtimeType]token expired or need re login!');
      } else {
        await init();
        res = await blocks[action]();
      }
    } else if (allows.containsKey(action)) {
      await init();
      try {
        res = await allows[action]();
      } catch (e,t) {
        res = Alm.error('[$runtimeType]action:$action error $e;');
        print(t);
      }
    } else {
      res = Alm.error('[$runtimeType] not found action:$action !!!');
    }
    cleanUp();
    return Alm.json(res);
  }

  void postMultiPartAllow(List<String> listFiles) async {
    var _listFiles = listFiles ?? [];
    var _fromDataDecode = (Map input,String key) {
      var map = <String,dynamic>{};
      input.forEach((k, value) {
        if(k==key){
          value.replaceAll('form-data; ', '').split(';').forEach((element) {
            var kv = element.split('=');
            map[kv.first.trim()] = kv.last.replaceAll('"', '').trim();
          });
        }else{
          map[k]=value;
        }
      });
      return map;
    };
    try {
      if (isBoundary) {
        var boundary = contentType.parameters['boundary'];
        await MimeMultipartTransformer(boundary).bind(request).forEach((element) async {
          try {
            if (Alm.isMap(element.headers,'content-disposition')) {
              var mp = _fromDataDecode(element.headers,'content-disposition');
              if (Alm.isMap(mp,'name')) {
                var kField = mp['name'];
                dynamic kVal;
                if (mp.length <= 2) {
                  kVal= await utf8.decoder.bind(element).join();
                } else {
                  kVal=mp;
                }
                if(postJson.containsKey(kField)){
                  if(postJson[kField] is List){
                    postJson[kField].add(kVal);
                  }else{
                    postJson[kField] =[postJson[kField],kVal];
                  }
                }else{
                  postJson[kField] = kVal;
                }
                if (kVal is Map&&_listFiles.contains(kField)) {
                  var bytes = <int>[];
                  (await element.toList()).forEach((e) => bytes.addAll(e));

                  if(postFiles.containsKey(kField)){
                    if(postFiles[kField] is List){
                      postFiles[kField].add(bytes);
                    }else{
                      postFiles[kField] =[postFiles[kField],bytes];
                    }
                  }else{
                    postFiles[kField] = bytes;
                  }

                }

              }
            }
          } catch (e) {
            log('postMultiPart.element ${element.headers} error $e');
          }
        });
      }
    } catch (e) {
      log('postMultiPart error $e');
    }
  }

  static void start(kRouteMethod routeMap, {int port = 4040, Future<void> onStart, Future<void> onClose, Function onUpdate, Map<String, String> responseHeaders}) async {
    var server = await HttpServer.bind('0.0.0.0', port);
    if (onStart != null) await onStart;

    log('Listening on http://${server.address.host}:${server.port}/');

    var _responseHeaders = responseHeaders ??
        {
          'Access-Control-Allow-Origin': '*',
          'Access-Control-Allow-Methods': '*',
          'Access-Control-Allow-Headers': '*',
        };

    await for (HttpRequest request in server) {
      if (onUpdate != null) onUpdate();

      var timeIn=Alm.timeint();

      var reqMsg = '${request.method} ${request.requestedUri}(${request.contentLength})';

      var ip ='${request.connectionInfo.remoteAddress.host}:${request.connectionInfo.remotePort}';

      var response = request.response;

      dynamic err;

      try {
        _responseHeaders.forEach((key, value) {
          response.headers.add(key, value);
        });
        response.headers.add('Server', 'Dart-Restful:1.0.0;');
        response.headers.add('Server-Info', 'Author:almpazel@gmail.com-Date:2020/2025;');
        dynamic resBody='';
        response.statusCode = HttpStatus.processing;
        try {
          response.headers.contentType = ContentType.json;
          var rs = routeMap(request);
          if (rs is Api) {
            resBody = await rs.enter(request);
          } else {
            resBody = rs;
          }
          response.statusCode = HttpStatus.ok;
        } catch (e,t) {
          log(t);
          err = e;
          response.statusCode = HttpStatus.internalServerError;
          resBody = Alm.json(Alm.error(e.toString()));
        }
         if (resBody is List) {
          response.add(resBody);
        } else{
          response.write(resBody);
        }

        await response.close();

        var timeDiff=Alm.timediff(timeIn);

        log('${Alm.timestamp()} $ip $reqMsg => ${response.statusCode} (${resBody.length}) ${err != null ? 'error:$err' : ''} :$timeDiff ');
      } catch (e) {
        log('error:$e');
      }
    }
    if (onClose != null) await onClose;
  }

  @override
  Map<String, kApiMethod> allows = {};

  @override
  Map<String, kApiMethod> blocks = {};

  @override
  HttpRequest request;

  @override
  String get headerToken => request?.headers?.value('token');

  @override
  Future<bool> headerCheckToken() async => true;

  void cleanUp() {
    postJson?.clear();
    postJson=null;
  }
}
